# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from . import base


class OdooUsers(base.OdooModule):
    _name = "Users"
    _key = "users"

    def apply(self):
        if self._key in self._datas:
            super(OdooUsers, self).apply()
            self.execute()


    def execute(self, datas=None):
        if not datas:
            datas = self._datas
        sections = datas.get(self._key, {})
        for section in sections:
            users = sections.get(section)
            self.logger.info("\t- %s" % section)
            for user in users:
                self.logger.info("\t\t- User %s" % user)
                self.update_user(users[user])
        if self._key in self._datas:
            self._datas.pop(self._key)
        if self._key in datas:
            datas.pop(self._key)

    def update_user(self, user_values):
        groups_id = []
        context = dict(user_values.get('context', {}))
        context.update(self._context)
        for group in user_values.get("groups_id", []):
            if group == "unlink all":
                groups_id.append((5,))
            else:
                groups_id.append(
                    (4, self._connection.get_ref(group)))

        login = user_values.get('login')
        if user_values.get('force_id', False):
            user_id = self.get_id_from_xml_id(user_values.get('force_id'))
        else:
            user_id = self.search('res.users', [('login', '=', login)], order='id', context=context)

        vals = dict(user_values.get('values', {}))
        if login:
            vals['login'] = login
        if groups_id:
            vals['groups_id'] = groups_id
        context['active_test'] = False
        if not user_id:
            self.execute_odoo('res.users', 'create', [vals], {'context': context})
            self.create_xml_id('external_config', 'res.users', login, user_id)
        else:
            self.execute_odoo('res.users', 'write', [user_id, vals], {'context': context})


# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import importlib.util as ilu
from . import imports
from ..utils import Utils as utils


class PythonScript(imports.OdooImports):
    _name = "PythonScript"
    _key = "python_script"
    auto_apply = False

    def get_func(self, data):
            file_path = data.get('file', False)
            method_name = data.get('method', False)
            self.logger.info("Python Script %s %s" % (file_path, method_name))
            path = utils.get_file_full_path(data.get('file'))
            spec = ilu.spec_from_file_location('import_specific', path)
            specific_lib = ilu.module_from_spec(spec)
            spec.loader.exec_module(specific_lib)
            return getattr(self._import_manager, method_name)

    def apply(self):
        self.prepare_extra_odoo_connections()
        self.prepare_sql_connections()
        datas = self._datas.get(self._key, {})
        for script_section in datas:
            if not self.install_mode() and datas[script_section].get('on_install_only', False):
                return
            script_data = datas[script_section]
            func = self.get_func(script_data)
            params = script_data.get('params', {})
            params['configurator'] = self._configurator
            params['config'] = self._configurator.config
            func(params=params)



# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import os
import pickle

from .logging import get_logger

logger = get_logger(__name__)

try:
    import psycopg
    from psycopg.rows import dict_row
    import pymssql
    import pyodbc
    import mysql.connector
except Exception as err:
    logger.error(err)
    pass


class SqlConnection:
    _name = "SQL"
    connection = None

    def __init__(self, db_type, url, dbname, username, password, port=5432, **kwargs):
        self.db_type = db_type
        self.url = url
        self.dbname = dbname
        self.username = username
        self.password = password
        self.port = port
        if self.db_type == "postgresql":
            self.connection = psycopg.connect(
                user=self.username,
                password=self.password,
                host=self.url,
                dbname=self.dbname,
                port=self.port,
                row_factory=dict_row
            )
        elif self.db_type == "mysql":
            self.connection = mysql.connector.connect(
                host=self.url,
                user=self.username,
                password=self.password,
                database=self.dbname
            )
        elif self.db_type == "mssql":
            driver = 'ODBC Driver 18 for SQL Server'
            connectionString = f'DRIVER={driver};SERVER={url};DATABASE={dbname};UID={username};Encrypt=no;PWD={password}'
            try:
                self.connection = pyodbc.connect(connectionString)
            except:
                print(
                    "https://learn.microsoft.com/en-us/sql/connect/odbc/linux-mac/installing-the-microsoft-odbc-driver-for-sql-server")

    def execute(self, query, cache=False):
        if cache:
            if not os.path.exists('.cache'):
                logger.info("Create cache directory .cache")
                os.makedirs('.cache')
            cache_filename = ".cache/" + cache + ".pickle"
            if os.path.exists(cache_filename):
                with open(cache_filename, "rb") as infile:
                    results = pickle.load(infile)
                return results
        cursor = self.connection.cursor()
        cursor.execute(query)

        columns = [column[0] for column in cursor.description]
        unique_columns = []
        column_count = {}
        for column in columns:
            if column in column_count:
                column_count[column] += 1
                unique_columns.append(f"{column}_{column_count[column]}")
            else:
                column_count[column] = 0
                unique_columns.append(column)

        results = []
        for row in cursor.fetchall():
            results.append(dict(zip(unique_columns, row)))
        cursor.close()
        if cache:
            with open(cache_filename, "wb") as outfile:
                pickle.dump(results, outfile)
        return results



# Copyright (C) 2023 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from .logging import get_logger
from s6r_bitwarden_cli import BitwardenCli
from .apps import connection

logger = get_logger("Bitwarden".ljust(20))


class Bitwarden:

    def __init__(self, configurator):
        self.configurator = configurator
        self.init_bitwarden_cli()

    def get_bitwarden_params(self):
        bitwarden_params = ['bitwarden_username', 'bitwarden_password', 'bitwarden_session_key']
        return {param: self.configurator.config.get(param) for param in bitwarden_params}

    def init_bitwarden_cli(self):
        params = self.get_bitwarden_params()
        connection.OdooConnection(self.configurator).pre_config(params)
        if params['bitwarden_session_key']:
            self.bw_cli = BitwardenCli()
            if not self.bw_cli.session_key:
                self.bw_cli.session_key = params['bitwarden_session_key']
        elif params['bitwarden_username'] and params['bitwarden_password']:
            self.bw_cli = BitwardenCli(params['bitwarden_username'], params['bitwarden_password'])
        else:
            self.bw_cli = BitwardenCli()

    def get_bitwarden_password(self, collection_name, name, default=''):
        try:
            return self.bw_cli.get_item_password(name, collection_name=collection_name)
        except Exception as err:
            logger.error(err)
            pass
            return default

    def get_bitwarden_username(self, collection_name, name):
        try:
            return self.bw_cli.get_item_username(name, collection_name=collection_name)
        except Exception as err:
            logger.error(err)
            pass
            return ''

    def get_bitwarden_field(self, collection_name, item_name, field_name):
        try:
            return self.bw_cli.get_item_field(item_name, field_name, collection_name=collection_name)
        except Exception as err:
            logger.error(err)
            pass
            return ''



# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from collections import OrderedDict
from . import base


class OdooConfig(base.OdooModule):
    _name = "Config"

    def __init__(self, configurator, auto_apply=True):
        self.auto_apply = auto_apply
        super(OdooConfig, self).__init__(configurator)

    def apply(self):
        super(OdooConfig, self).apply()
        config_values = self.prepare_config_values()
        for key in config_values:
            self.logger.info("\tConfig - %s" % key)
            self.execute_config(config_values[key])

    def execute_script_config(self, datas=False):
        datas = datas if datas else self._datas
        for key in datas:
            if isinstance(datas.get(key), dict) or isinstance(datas.get(key), OrderedDict):
                if datas.get(key).get('config', {}):
                    self.logger.info("\tConfig - %s" % key)
                    values = self.prepare_company_values(datas.get(key).get('config', {}))
                    self.execute_config(values)

    def prepare_config_values(self, datas=False):
        datas = datas if datas else self._datas
        config_values = {}
        for key in datas:
            config_data = datas.get(key)
            if isinstance(config_data, dict) or isinstance(config_data, OrderedDict):
                if config_data.get('config', {}):
                    config_values[key] = config_data.get('config')
        if config_values:
            self.prepare_company_values(datas)
        return config_values

    def prepare_company_values(self, values):
        self.pre_config(values)
        if values.get('company_id'):
            self._context['allowed_company_ids'] = [values.get('company_id')]
        if values.get('allowed_company_ids'):
            self._context['allowed_company_ids'] = values.get('allowed_company_ids')
            values.pop('allowed_company_ids')
        return values

    def execute_config(self, config):
        context = self._context
        for key in config:
            if isinstance(config[key], str) and config[key].startswith('get_'):
                config[key] = self.safe_eval(config[key])
        domain = []

        if 'context' in config:
            context.update(config.pop('context'))

        if 'company_id' in config:
            domain.append(('company_id', '=', config['company_id']))
            context['allowed_company_ids'] = [config['company_id']]

        config_id = self.execute_odoo('res.config.settings', 'create', [self.deep_convert_dict(config)],
                                      {'context': context})
        self.execute_odoo('res.config.settings', 'execute', [config_id], {'context': context})
        self.logger.info("\t\t * Done")



# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from s6r_odoo import OdooConnection as Orm
from collections import OrderedDict
import importlib.util as ilu
from . import base
from ..utils import Utils as utils
from ..sql import SqlConnection


class OdooImports(base.OdooModule):
    _name = "Imports"

    def __init__(self, configurator, auto_apply=True):
        self.auto_apply = auto_apply
        super().__init__(configurator)

    def get_func(self, data):
        if data.get('specific_import', False):
            if data.get('file_path', ''):
                self.logger.info("Specific import %s %s : %s" % (data.get('specific_import', False),
                                                                 data.get('specific_method', False),
                                                                 data.get('file_path', '')))
            else:
                self.logger.info("Specific import %s %s" % (data.get('specific_import', False),
                                                            data.get('specific_method', False)))
            path = utils.get_file_full_path(data.get('specific_import'))
            spec = ilu.spec_from_file_location('import_specific', path)
            specific_lib = ilu.module_from_spec(spec)
            spec.loader.exec_module(specific_lib)
            func = getattr(self._import_manager, data.get('specific_method', False))
        else:
            func = getattr(self._import_manager, 'import_csv')
        return func

    def apply(self, datas=None):
        super(OdooImports, self).apply()
        self.prepare_extra_odoo_connections()
        self.prepare_sql_connections()
        script_datas = datas or self._datas
        if self._import_manager and not hasattr(self._import_manager, '_datas'):
            self._import_manager._datas = self._datas

        import_datas = script_datas.get('import_data', {})
        if import_datas:
            self.logger.info("\t- %s" % "Global")
        for import_name in import_datas:
            if not self.install_mode() and import_datas[import_name].get('on_install_only', False):
                return
            import_data = import_datas[import_name]
            func = self.get_func(import_data)
            func(utils.get_file_full_path(import_data.get('file_path', '')),
                 import_data.get('model', ''), params=import_data)
        for key in script_datas:
            if isinstance(script_datas.get(key), dict) or isinstance(script_datas.get(key), OrderedDict):
                for key_import, import_data in script_datas.get(key, {}).get('import_data', {}).items():
                    self.logger.info("\t- %s" % key_import)
                    if not self.install_mode() and import_data.get('on_install_only', False):
                        self.logger.info("\t\t* skipped not install mode")
                        return
                    func = self.get_func(import_data)
                    func(utils.get_file_full_path(import_data.get('file_path', '')),
                         import_data.get('model', ''), params=import_data)

    def prepare_extra_odoo_connections(self):
        auth = self._datas.get('auth', {})
        if auth:
            for key in auth:
                if key != 'odoo' and not hasattr(self._import_manager, key):
                    auth_values = auth.get(key, {})
                    try:
                        setattr(self._import_manager, key, Orm(auth_values['url'],
                                               auth_values['dbname'],
                                               auth_values['username'],
                                               auth_values['password'],
                                               version=auth_values.get('version', False),
                                               ))
                    except ConnectionError:
                        exit(1)


    def prepare_sql_connections(self):
        sql_auth = self._datas.get('sql_auth', {})
        if sql_auth:
            for key in sql_auth:
                if not hasattr(self._import_manager, key):
                    auth_values = sql_auth.get(key, {})
                    try:
                        setattr(self._import_manager, key, SqlConnection(**auth_values))
                    except Exception as err:
                        self.logger.error(err)
                        exit(1)


# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from . import base


class OdooConnection(base.OdooModule):
    _name = "Auth"
    _key = "auth"

    def apply(self):
        pass
